; ZZT music player for Famicom (v0.1)
; This file is public domain

; Instructions: Concatenate the ".zzm" file after "zztmusicplayer.nes" and
; name the resulting file with a ".nsf" extension. It must consist of one
; song! (If you have more than one, it will play all of them without
; stopping; this program does not understand ZZM commands.)


; === Macros ===
assert	.macro
	.if \1
	.else
	.fail "Assertion failed: \1"
	.endif
	.endm


; === RAM variables ===
	.zp
playlo	.ds 1 ; low byte of play address
playhi	.ds 1 ; high byte of play address
playb	.ds 1 ; bank number of play address
nlen	.ds 1 ; note length (default="T")
octave	.ds 1 ; current octave (0 to 6, default=3)
ticks	.ds 1 ; number of ticks remaining for note


; === Data of .NSF header ===
	.code
	.bank 0
	.org $0000
	.db "NESM", $1A, $01, $01, $01 ; version, total, starting
	.dw $8000, s_init, s_play ; load/init/play address
	.db "<?>", $00 ; title
	.ds 28
	.db "<?>", $00 ; composer
	.ds 28
	.db "<?>", $00 ; maker
	.ds 28
	.dw 16640 ; play speed (NTSC)
	.db $00, $01, $00, $00, $00, $00, $00, $00 ; bank init
	.dw 19997 ; play speed (PAL)
	.db $00, $01, $00, $00, $00, $00 ; TV, chip, 4x reserved


; === Program ===
	; ZZT #PLAY subcommand subroutines
	.bank 1
	.org $8000

	; reserved point
	.ds 1

	; notes: ABCDEFG
c_note	and #$1F
	tax
	lda <playb ; select bank for looh ahead
	sta $5FFA
	ldy #0
	lda [playlo],y ; look ahead
	tay
	lda nldata,x ; use alphabetic lookup table
	cpy #'!' ; flat
	bne c_note1
	sbc #1 ; carry flag is set
c_note1	cpy #'#' ; sharp
	bne c_note2
	adc #0 ; carry flag is set
c_note2	asl a ; multiply by 2 to use frequency table
	tax   ; (entries are 2 bytes long)
	lda freqd,x ; low byte
	sta <ticks ; reuse <ticks temporarily
	inx
	lda freqd,x ; high byte
	ldy <octave
c_note3	lsr a ; shift right by octave up
	ror <ticks ; shift right by octave up
	dey
	bpl c_note3
	dec <ticks ; subtract 1
	ora #$80
	sta $9002 ; VRC6 square wave period high
	lda <ticks
	sta $9001 ; VRC6 square wave period low

	; rest: X
c_rest	lda <nlen
	sta <ticks
	rts

	; percussion: 012456789
c_perc	adc #((dmcdata&$FFF)>>6)
	sta $4012 ; DPCM sample address
	lda #$40
	sta $4011 ; DPCM direct load
	lda #$04
	sta $4013 ; DPCM sample length
	lda #$10
	sta $4015 ; enable DPCM
	bne c_rest

	; length: WHQIST
c_len	and #$1F
	tax
	lda nldata,x
	sta <nlen
	bne c_nop1

	; octave down: -
c_octdn	lda <octave
	beq c_nop1
	dec <octave
	bpl c_nop1

	; octave down: +
c_octup	lda <octave
	cmp #$06
	beq c_nop1
	inc <octave
	bpl c_nop1

	; increase length by 50%: .
c_dot	lda <nlen
	lsr a
	adc <nlen
	sta <nlen
c_nop1	bcc c_nop

	; divide length by three: 3
c_tri	lda <nlen
	lsr a
	lsr a
	adc <nlen
	ror a
	lsr a
	adc <nlen
	ror a
	lsr a
	adc <nlen
	ror a
	lsr a
	adc <nlen
	ror a
	lsr a
	sta <nlen
	bpl c_nop

	; carriage return/line feed
c_crlf	lda #3
	sta <octave
	lda #1
	sta <nlen
	bpl c_nop

	; end of file
c_eof	beq c_eof

	; comments: ;
c_com	jsr getch
	cmp #$0A
	bne c_com
	assert *<$80FF

	; no operation; read next command
c_nop	jsr getch
	lda #$80
	pha
	lda transl,x
	pha
	txa
	rts

	; alignment to 64 bytes
	.org *+((0-*)&$3F)

	; Translation table
transl	.ds 128

	; DPCM samples (64 bytes each)
dmcdata	.incbin "zztperc/0.bin"
	.incbin "zztperc/1.bin"
	.incbin "zztperc/2.bin"
	; data table in middle of DPCM
nldata	.db 0,9,11,0,2,4,5,7   ; .ABCDEFG
	.db 48,12,0,0,0,0,0,0  ; HIJKLMNO
	.db 0,24,0,6,3,0,0,96  ; PQRSTUVW
	.db 0,0,0,0,0,0,0,0    ; XYZ.....
freqd	.dw $1AB9,$1939,$17CF,$1679
	.dw $1536,$1405,$12E5,$11D6
	.dw $10D6,$0FE4,$0EFF,$0E28
	.dw $0000,$0000,$0000,$0000
	; more DPCM samples
	.incbin "zztperc/4.bin"
	.incbin "zztperc/5.bin"
	.incbin "zztperc/6.bin"
	.incbin "zztperc/7.bin"
	.incbin "zztperc/8.bin"
	.incbin "zztperc/9.bin"

	; Init
s_init	sta <playb ; A is already zero
	sta $4015 ; status (disable all channels)
	sta $9003 ; VRC6 frequency control
	lda #high(s_eof+$2000) ; start at $Axxx
	sta <playhi
	lda #low(s_eof+$2000)
	sta <playlo
	lda #$7F
	sta $9000 ; VRC6 square duty/volume
	lsr a
	sta $4010 ; DPCM flag/frequency
	jmp c_crlf

	; Play each frame
s_play	ldx <ticks
	beq s_play0
	dex
	stx <ticks
	bne s_play1
	stx $4015 ; turn off DPCM (X=0)
	stx $9002 ; turn off VRC6 square wave
s_play0	jmp c_nop
s_play1	rts

	; Get next character
getch	lda <playb
	sta $5FFA
	ldy #0
	lda [playlo],y
	tax
	lda <playlo
	clc
	adc #1
	sta <playlo
	lda <playhi
	adc #0
	cmp #$B0
	and #$AF
	sta <playhi
	lda <playb
	adc #0
	sta <playb
	txa
	rts

	; End of file
	assert *<$8FFE
s_eof	.ds 0

; === Write file ===
	.emu X
	.org $FFFC
	.dw $8000 ; reset vector

	.org $8000

	; Write .NSF header
	lda #$00
	sta $2000 ; bank number
	sta $2007 ; bank output length high byte
	lda #$80
	sta $2006 ; bank output length low byte
	sta $2008 ; write bank to output file

	; Make translation table
	lda #$01
	sta $2000 ; bank number

	; Fill with (c_nop) (except first entry)
	ldy #$00
	lda #low(c_eof-1)
mkttd0	sta transl-$4000,y
	lda #low(c_nop-1)
	iny
	cpy #$80
	bne mkttd0

	; Add data
	ldx #$FF
mkttd1	inx
	cpx #ttd_end
	beq wrrom
	lda <0,x
	clc
	sbc #0
mkttd2	inx
	ldy <0,x
	beq mkttd1
	sta transl-$4000,y
	jmp mkttd2

	; Write program ROM
wrrom	lda #high(s_eof)-$80
	sta $2007 ; bank output length high byte
	lda #low(s_eof)
	sta $2006 ; bank output length low byte
	sta $2008 ; write bank to output file

	hlt

	; Translation table data
	.org $0000
	.db low(c_note),"ABCDEFGabcdefg",0
	.db low(c_rest),"Xx",0
	.db low(c_len),"WHQISTwhqist",0
	.db low(c_octdn),"-",0
	.db low(c_octup),"+",0
	.db low(c_dot),".",0
	.db low(c_tri),"3",0
	.db low(c_perc),"012456789",0
	.db low(c_crlf),"\r\n",0
	.db low(c_com),";",0
ttd_end	.ds 0

	.code
